#!/bin/false
# -*- coding: utf-8 -*-

"""
This module provides unit tests for `ptutils.punning`.
"""
import pytest
import math
from ptutils.encoding import HAVE_YAML
from ptutils.punning import to_string, boolify, listify, as_numeric, as_encoded, as_literal, canonize
from ptutils.undefined import UNDEFINED


class C:
    pass


c = C()


TO_STRING_TEST_CASES = [
    (b"hello", "hello"),
    ("hello", "hello"),
    (5, "5"),
    (None, "None"),
    (False, "False"),
    (True, "True"),
    ([True, 1], "[True, 1]"),
    (list(), "[]"),
    ({"a": True, "b": 1}, "{'a': True, 'b': 1}"),
    (dict(), "{}"),
    (c, str(c))
]


@pytest.mark.parametrize(
    argnames  = ["obj", "expected"],
    argvalues = TO_STRING_TEST_CASES,
    ids       = [f"{repr(x[0])}->{repr(x[1])}" for x in TO_STRING_TEST_CASES]
)
def test_to_string(obj, expected):
    if isinstance(expected, type) and issubclass(expected, Exception):
        with pytest.raises(expected) as _:
            to_string(obj)
    else:
        assert to_string(obj) == expected


BOOLIFY_TEST_CASES = [
    # normal cases
    (False, False),
    (True, True),
    (None, False),
    (UNDEFINED, False),
    ("yes", True),
    ("on", True),
    ("enabled", True),
    ("true", True),
    ("YES", True),
    ("ON", True),
    ("EnAbLeD", True),
    ("trUE", True),
    ("1", True),
    ("no", False),
    ("disabled", False),
    ("off", False),
    ("false", False),
    ("0", False),
    # error cases
    (b"hello", TypeError),
    ("hello", TypeError),
    (5, TypeError),
    ([True, 1], TypeError),
    (list(), TypeError),
    ({"a": True, "b": 1}, TypeError),
    (dict(), TypeError),
]


@pytest.mark.parametrize(
    argnames  = ["obj", "expected"],
    argvalues = BOOLIFY_TEST_CASES,
    ids       = [f"{repr(x[0])}->{repr(x[1])}" for x in BOOLIFY_TEST_CASES]
)
def test_boolify(obj, expected):
    if isinstance(expected, type) and issubclass(expected, Exception):
        with pytest.raises(expected) as _:
            boolify(obj)
    else:
        assert boolify(obj) == expected


LISTIFY_TEST_CASES = [
    (None,               False, [None]),
    ("a",                False, ["a"]),
    (1,                  False, [1]),
    (1.5,                False, [1.5]),
    (False,              False, [False]),
    (c,                  False, [c]),
    ({'a': 1, 'b': 'B'}, False, [{'a': 1, 'b': 'B'}]),
    ([1, 2, 3],          False, [1, 2, 3]),
    (set([1, 2, 3]),     False, set([1, 2, 3])),
    (tuple([1, 2, 3]),   False, tuple([1, 2, 3])),
    (None,               True, [None]),
    ("a",                True, ["a"]),
    (1,                  True, [1]),
    (1.5,                True, [1.5]),
    (False,              True, [False]),
    (c,                  True, [c]),
    ({'a': 1, 'b': 'B'}, True, [{'a': 1, 'b': 'B'}]),
    ([1, 2, 3],          True, [1, 2, 3]),
    (set([1, 2, 3]),     True, [1, 2, 3]),
    (tuple([1, 2, 3]),   True, [1, 2, 3])
]


@pytest.mark.parametrize(
    argnames  = ["obj", "strict", "expected"],
    argvalues = LISTIFY_TEST_CASES,
    ids       = [f"{repr(x[0])}->{repr(x[1])}" for x in LISTIFY_TEST_CASES]
)
def test_listify(obj, strict, expected):
    assert listify(obj, strict=strict) == expected


CANONIZE_TEST_CASES = [
    (None,                    False, None),
    ("a",                     False, "a"),
    (1,                       False, 1),
    (1.5,                     False, 1.5),
    (False,                   False, False),
    (c,                       False, c),
    ({'a': 1, 'b': 'B'},      False, {'a': 1, 'b': 'B'}),
    ([1, 2, 3],               False, [1, 2, 3]),
    (set([1, 2, 3]),          False, {1, 2, 3}),
    (tuple([1, 2, 3]),        False, tuple([1, 2, 3])),
    ("None",                  False, None),
    ("null",                  False, None),
    ("'None'",                False, None),
    ("'null'",                False, None),
    ('"a"',                   False, "a"),
    ("1",                     False, 1),
    ("1.5",                   False, 1.5),
    ("False",                 False, False),
    ('{"a": 1, "b": "B"}',    False, {'a': 1, 'b': 'B'}),
    ("[1, 2, 3]",             False, [1, 2, 3]),
    ("---\n",                 True, None),
    ("---\n- 1\n- 2\n- 3",    True, [1, 2, 3]),
    ("---\na: 1\nb: 2\nc: 3", True, {'a': 1, 'b': 2, 'c': 3}),
    ("this shouldn't parse",  False, "this shouldn't parse"),
    (b'{"a": 1, "b": "B"}',   False, {'a': 1, 'b': 'B'}),
    (b'true',                 False, True),
    (b'fAlsE',                False, False),
    ("---\ninvalid yaml: : :- ::", True, "---\ninvalid yaml: : :- ::"),

]


@pytest.mark.parametrize(
    argnames  = ["obj", "expect_yaml", "expected"],
    argvalues = CANONIZE_TEST_CASES,
    ids       = [f"{repr(x[0])}->{repr(x[1])}" for x in CANONIZE_TEST_CASES]
)
def test_canonize(obj, expect_yaml, expected):
    if expect_yaml and not HAVE_YAML:
        pytest.skip("YAML support unavailable on this platform")

    assert canonize(obj) == expected


AS_NUMERIC_TEST_CASES = [
    (None,                    TypeError),
    ("a",                     ValueError),
    ("+inf",                  float("+inf")),
    ("-inf",                  float("-inf")),
    ("nan",                   float("nan")),
    (1,                       1),
    (1.5,                     1.5),
    ("1",                     1),
    ("1.5",                   1.5),
]


@pytest.mark.parametrize(
    argnames  = ["obj", "expected"],
    argvalues = AS_NUMERIC_TEST_CASES,
    ids       = [f"{repr(x[0])}->{repr(x[1])}" for x in AS_NUMERIC_TEST_CASES]
)
def test_as_numeric(obj, expected):
    if isinstance(expected, type) and issubclass(expected, Exception):
        with pytest.raises(expected) as _:
            as_numeric(obj)
    else:
        v = as_numeric(obj)
        if math.isnan(v):
            assert math.isnan(expected)

        elif math.isinf(v):
            assert math.isinf(expected)

        else:
            assert v == expected



#!/bin/false
# -*- coding: utf-8 -*-

"""This module provides unit tests for `ptutils.file`."""

import os
import sys
import pytest
import pathlib
from ptutils.file import ( Path, File, Folder, FolderSet )
from ptutils.undefined import UNDEFINED, is_defined
from ptutils.encoding import decode_json, decode_yaml, encode_json
from ptutils.io import load_txt, load_json, load_yaml, save_txt, save_json, save_yaml
from ._fs_fixture import sample_file_structure  # noqa: F401
from ._fs_fixture import FILE_STRUCTURE
from ._fs_fixture import localize_path

PATH_EXPECTATIONS_POSIX = {
    "/a/path/to/a/file.ext": {
        "parent":       Path("/a/path/to/a"),
        "parts":        tuple(["/", "a", "path", "to", "a", "file.ext"]),
        "drive":        "",
        "root":         "/",
        "anchor":       "/",
        "name":         "file.ext",
        "suffix":       ".ext",
        "suffixes":     [".ext"],
        "stem":         "file",
        "exists":       False,
        "as_posix":     "/a/path/to/a/file.ext",
        "as_uri":       "file:///a/path/to/a/file.ext",
        "is_absolute":  True,
        "is_reserved":  False,
        "expanduser":   "/a/path/to/a/file.ext"
    },
    "a/path/to/a/file.ext1.ext2": {
        "parent":       Path("a/path/to/a"),
        "parts":        tuple("a/path/to/a/file.ext1.ext2".split("/")),
        "drive":        "",
        "root":         "",
        "anchor":       "",
        "name":         "file.ext1.ext2",
        "suffix":       ".ext2",
        "suffixes":     [".ext1", ".ext2"],
        "stem":         "file.ext1",
        "exists":       False,
        "as_posix":     "a/path/to/a/file.ext1.ext2",
        "as_uri":       ValueError,
        "is_absolute":  False,
        "is_reserved":  False,
        "expanduser":   "a/path/to/a/file.ext1.ext2"
    }
}
PATH_EXPECTATION_KEYS_POSIX = [
    "parent",
    "parts",
    "drive",
    "root",
    "anchor",
    "name",
    "suffix",
    "suffixes",
    "stem",
    "exists",
    "as_posix",
    "as_uri",
    "is_absolute",
    "is_reserved",
    "expanduser"
]


@pytest.mark.parametrize(
    argnames  = "key",
    argvalues = PATH_EXPECTATION_KEYS_POSIX,
    ids       = PATH_EXPECTATION_KEYS_POSIX
)
@pytest.mark.parametrize(
    argnames  = ["path", "expectations"],
    argvalues = PATH_EXPECTATIONS_POSIX.items(),
    ids       = [ x[0] for x in PATH_EXPECTATIONS_POSIX.items() ]
)
def test_path(path, expectations, key):
    p = Path(path)
    e = expectations[key]
    if (isinstance(e, type)) and (issubclass(e, Exception)):
        # expecting an exception to be raised
        with pytest.raises(e):
            a = getattr(p, key, UNDEFINED)
            if callable(a):
                a()
    else:
        # not expecting an exception to be raised
        a = getattr(p, key, UNDEFINED)
        assert is_defined(a)
        if callable(a):
            a = a()
        assert a == expectations[key]


def test_path_wrapper_methods(sample_file_structure):  # noqa: F811
    folder = Path(sample_file_structure)
    folder2 = Path(sample_file_structure)

    # test context entry/exit
    with folder as _:

        # test is_xxx
        assert folder.is_absolute() is True
        assert folder.is_dir() is True
        assert folder.is_file() is False
        assert folder.is_symlink() is False
        assert folder.is_block_device() is False
        assert folder.is_char_device() is False
        assert folder.is_fifo() is False
        assert folder.is_socket() is False

        # test same file
        assert folder.samefile(folder2)

        # test iterdir
        count = sum(1 for _ in folder.iterdir())
        assert count > 0

        # test glob
        count = sum(1 for _ in folder.glob('*'))
        assert count > 0

        # test rglob
        count = sum(1 for _ in folder.rglob('*.json'))
        assert count > 0

        # test absolute
        assert isinstance(folder.absolute(), Path)

        # test resolve
        assert isinstance(folder.resolve(), Path)
        assert isinstance(folder.resolve(strict=True), Path)

        # test stat
        assert folder.stat() is not None

        # test owner/group
        if os.name != 'nt':
            assert folder.owner() is not None
            assert folder.group() is not None

        # test bytes
        assert bytes(folder) == str(folder).encode('utf-8')

        # test comparisons
        for part in ['abc', 'abc/def', 'abc/def/', '/abc', '/abd/def', '/abd/def/']:
            for case in [part, Path(part), pathlib.Path(part)]:
                assert isinstance(folder > case, bool)
                assert isinstance(folder < case, bool)
                assert isinstance(folder >= case, bool)
                assert isinstance(folder <= case, bool)

    # file write_text/read_text/unlink text
    content = encode_json(SAMPLE_DATAFILE_DATA)
    tmpfile = folder / "temp.file.json"
    with tmpfile.open("w") as _:
        tmpfile.write_text(content)
    assert tmpfile.exists()
    assert tmpfile.is_file()
    with tmpfile.open("r") as _:
        assert tmpfile.read_text() == content
    tmpfile.unlink()
    assert not tmpfile.exists()

    # file write_bytes/read_bytes/unlink binary
    content = encode_json(SAMPLE_DATAFILE_DATA).encode('utf-8')
    with tmpfile.open("wb") as _:
        tmpfile.write_bytes(content)
    assert tmpfile.exists()
    assert tmpfile.is_file()
    with tmpfile.open("rb") as _:
        assert tmpfile.read_bytes() == content
    tmpfile.unlink()
    assert not tmpfile.exists()

    # file touch/is_file/unlink
    tmpfile.touch()
    assert tmpfile.exists()
    tmpfile.unlink()
    assert not tmpfile.exists()

    # folder mkdir/rmdir
    folder3 = folder / "new-folder"
    assert not folder3.exists()
    folder3.mkdir()
    assert folder3.exists()
    assert folder3.is_dir()
    folder3.rmdir()
    assert not folder3.exists()

    # folder mkdir/rename/rmdir
    folder4 = folder / "new-folder.bak"
    assert not folder4.exists()
    folder3.mkdir()
    folder3.rename(folder4)
    assert folder4.exists()
    assert folder4.is_dir()
    folder4.rmdir()
    assert not folder4.exists()

    # file stat/lstat
    assert not tmpfile.exists()
    tmpfile.touch()
    assert tmpfile.exists()
    assert tmpfile.lstat() is not None
    assert tmpfile.stat() is not None
    assert tmpfile.stat() == tmpfile.lstat()
    tmpfile.unlink()
    assert not tmpfile.exists()

    if os.name != 'nt':
        assert not tmpfile.exists()
        tmpfile.touch()
        assert tmpfile.exists()
        tmpfile.chmod(0o444)
        assert (tmpfile.stat().st_mode & 0o777) == 0o444
        tmpfile.chmod(0o456)
        assert (tmpfile.stat().st_mode & 0o777) == 0o456
        tmpfile.unlink()
        assert not tmpfile.exists()

    # home/cwd class methods
    assert Path.cwd() == Path(os.getcwd())
    assert isinstance(Path.home(), Path)

    # folder symlink symlink_to/lstat/lchmod
    folder_symlink = folder / "folder_symlink"
    assert not folder_symlink.exists()
    folder_symlink.symlink_to(folder / "folder_symlink_target_dne", target_is_directory=True)
    assert folder_symlink.is_symlink()
    assert not folder_symlink.exists()
    assert folder_symlink.lstat() is not None
    with pytest.raises(Exception) as _:
        folder_symlink.stat()
    if os.name != 'nt':
        try:
            folder_symlink.lchmod(0o444)
            assert (folder_symlink.lstat().st_mode & 0o777) == 0o444
            folder_symlink.lchmod(0o456)
            assert (folder_symlink.lstat().st_mode & 0o777) == 0o456
        except NotImplementedError:
            pass
    folder_symlink.unlink()
    assert not folder_symlink.exists()
    assert not folder_symlink.is_symlink()

    # link stat/lstat
    # file_symlink   = folder / "file_symlink"
    # assert not file_symlink.exists()
    file_symlink = folder / "file_symlink"
    assert not file_symlink.exists()
    file_symlink.symlink_to(folder / "file_symlink_target_dne", target_is_directory=False)
    assert file_symlink.is_symlink()
    assert not file_symlink.exists()
    assert file_symlink.lstat() is not None
    with pytest.raises(Exception) as _:
        file_symlink.stat()
    if os.name != 'nt':
        try:
            file_symlink.lchmod(0o444)
            assert (file_symlink.lstat().st_mode & 0o777) == 0o444
            file_symlink.lchmod(0o456)
            assert (file_symlink.lstat().st_mode & 0o777) == 0o456
        except NotImplementedError:
            pass
    file_symlink.unlink()
    assert not file_symlink.exists()
    assert not file_symlink.is_symlink()

    assert isinstance(hash(folder), int)
    assert isinstance((folder != ''), bool)
    assert repr(folder) == f"Path('{str(folder)}')"
    assert (folder / "abc").with_name("xyz").name == "xyz"
    assert (folder / "abc").with_suffix(".xyz").suffix == ".xyz"
    assert folder.joinpath("abc", "xyz") == folder / "abc" / "xyz"
    assert ("x" / folder) == Path(os.path.join("x", sample_file_structure))
    assert len(folder.parents) == 2
    assert isinstance(folder.match('abc'), bool)
    assert (folder / "abc" / "xyz").relative_to(folder) == Path('abc/xyz')

    tmpfile1 = folder.joinpath('boring_things', 'thing.json')
    tmpfile2 = folder.joinpath('boring_things', 'another-thing.json')
    assert tmpfile1.exists()
    assert not tmpfile2.exists()
    tmpfile3 = tmpfile1.replace(tmpfile2)
    if sys.version_info >= (3, 8):
        assert tmpfile3 == tmpfile2
    else:
        assert tmpfile3 is None
    assert not tmpfile1.exists()
    assert tmpfile2.exists()
    tmpfile4 = tmpfile2.replace(tmpfile1)
    if sys.version_info >= (3, 8):
        assert tmpfile4 == tmpfile1
    else:
        assert tmpfile4 is None
    assert tmpfile1.exists()
    assert not tmpfile2.exists()
    if sys.version_info >= (3, 8):
        assert not tmpfile3.exists()
        assert tmpfile4.exists()


def test_path_static_methods():
    assert Path.path_of('abc') == Path('abc')
    assert Path.path_of(Path('abc')) == Path('abc')
    assert Path.path_of(pathlib.Path('abc')) == Path('abc')

    paths = [
        Path.path_of('abc'),
        Path.path_of(Path('abc')),
        Path.path_of(pathlib.Path('abc'))
    ]
    for path in paths:
        assert path.name == 'abc'

    assert list(Path.names_of(paths)) == paths
    folder = Path(localize_path('a/b/c/d/e'))
    ref    = Path(localize_path('a/b'))
    paths = [ (folder / x) for x in "vwxyz" ]
    assert list(Path.relative_paths_to(paths, ref)) == [ (Path(localize_path("c/d/e")) / x) for x in "vwxyz" ]


FILE_TEST_CASES = [
    ('boring_things/thing.json', 'json', True),
    ('boring_things/thing.yaml', 'yaml', True),
    ('boring_things/text_thing', 'text', True),
    ('boring_things/tbd_things/thing.json', 'json', False),
    ('boring_things/tbd_things/thing.yaml', 'yaml', False),
    ('boring_things/tbd_things/text_thing', 'text', False)
]


SAMPLE_TEXTFILE_DATA = '\n'.join([
    "This is a text file",
    "It has a few lines.",
    "There's nothing special about it.",
    "It can be used for testing",
    "text file manipulation ",
    "functions and classes."
])


SAMPLE_DATAFILE_DATA = {
    "a": True,
    "b": 12345,
    "c": None,
    "d": 3.14,
    "e": [False, 678, None, 1.2, [], {}],
    "f": {
        "1": 1,
        "2": 2.0,
        "3": None,
        "4": [False, 678, None, 1.2, [], {}],
        "5": {}
    }
}


LOADERS = {
    "text": load_txt,
    "json": load_json,
    "yaml": load_yaml
}


SAVERS = {
    "text": save_txt,
    "json": save_json,
    "yaml": save_yaml
}


@pytest.mark.parametrize(
    argnames  = ["path", "format", "exists"],
    argvalues = FILE_TEST_CASES,
    ids       = [f"{x[0]}(format={x[1]},exists={x[2]})" for x in FILE_TEST_CASES]
)
def test_file(sample_file_structure, path, format, exists):  # noqa: F811
    path = os.path.join(sample_file_structure, localize_path(path))
    file = File(path)
    assert file.exists() == exists
    assert file.loader   is LOADERS[format]  # noqa: E272,E271
    assert file.saver    is SAVERS[format]   # noqa: E272,E271
    assert file.parent   == os.path.join(sample_file_structure, os.path.dirname(path))

    if exists:
        dir_data = FILE_STRUCTURE['boring_things']
        if format == "text":
            assert file.content == dir_data['text_thing']

        else:
            if format == "json":
                assert file.content == decode_json(dir_data['thing.json'])
                assert list(file.lines) == dir_data['thing.json'].split('\n')
            else:
                assert file.content == decode_yaml(dir_data['thing.yaml'])
                assert list(file.lines) == dir_data['thing.yaml'].split('\n')

    else:
        if format == "text":
            content = SAMPLE_TEXTFILE_DATA
        else:
            content = SAMPLE_DATAFILE_DATA

        assert not file.exists()
        file.initialize_content(content = content)
        assert file.exists()
        assert file.content == content
        file.initialize_content(content = content)
        assert file.exists()
        assert file.content == content
        file.initialize_content(content = UNDEFINED)
        assert file.exists()
        assert file.content == content


def test_folder(sample_file_structure):  # noqa: F811
    folder = Folder(sample_file_structure)
    subfiles = [ File(folder / x) for x in ["a_file", "another_file"] ]
    subdirs  = [ Folder(folder / x) for x in ["boring_things", "interesting_things", "apples"] ]
    assert folder.exists()
    assert set(list(folder.search_files())) == set(subfiles)
    assert set(list(folder.search_subdirectories())) == set(subdirs)
    assert set(folder.subdirectories) == set(subdirs)
    assert set(folder.files) == set(subfiles)
    assert folder.child("abc") == folder / "abc"
    assert isinstance(folder.child_file("abc"), File)
    assert folder.child_file("abc") == File(folder / "abc")
    assert isinstance(folder.child_folder("abc"), Folder)
    assert folder.child_folder("abc") == Folder(folder / "abc")


def test_folderset(sample_file_structure):  # noqa: F811
    folderset = FolderSet(sample_file_structure)
    assert folderset.folder == Folder(sample_file_structure)
    f1 = folderset.a_file
    assert isinstance(f1, File)
    assert f1.exists()
    with pytest.raises(AttributeError):
        folderset.a_file_that_doesnt_exist
    with pytest.raises(KeyError):
        folderset['a_file_that_doesnt_exist']
    with pytest.raises(KeyError):
        folderset / 'a_file_that_doesnt_exist'
    f3 = folderset.apples
    assert isinstance(f3, FolderSet)
    assert f3.folder.exists()
    assert f3.folder == folderset.folder / 'apples'
    assert (folderset / 'apples').folder == folderset.folder / 'apples'
    with pytest.raises(KeyError):
        folderset['a_file_that_doesnt_exist']
    assert folderset['$$a_file'] == folderset.a_file


def test_folderset_structured(sample_file_structure):  # noqa: F811
    folderset = FolderSet(
        sample_file_structure,
        structure={
            "$$files": [
                "a_file"
            ],
            "$$folders": [
                "apples"
            ],
            "boring_things": {}
        }
    )
    f1 = folderset.a_file
    assert isinstance(f1, File)
    assert f1.exists()

    with pytest.raises(AttributeError):
        folderset.a_file_that_doesnt_exist

    with pytest.raises(KeyError):
        folderset / 'a_file_that_doesnt_exist'

    with pytest.raises(KeyError):
        folderset['a_file_that_doesnt_exist']

    with pytest.raises(AttributeError):
        folderset.another_file

    with pytest.raises(KeyError):
        folderset["another_file"]

    with pytest.raises(KeyError):
        folderset / "another_file"

    f2 = folderset.apples
    assert isinstance(f2, FolderSet)
    assert f2.folder.exists()
    assert f2.folder == folderset.folder / 'apples'
    assert (folderset / 'apples').folder == folderset.folder / 'apples'

    f3 = folderset.boring_things
    assert isinstance(f3, FolderSet)
    assert f3.folder.exists()
    assert f3.folder == folderset.folder / 'boring_things'
    assert (folderset / 'boring_things').folder == folderset.folder / 'boring_things'

    with pytest.raises(AttributeError):
        folderset.interesting_things

    with pytest.raises(KeyError):
        folderset["interesting_things"]

    with pytest.raises(KeyError):
        folderset / "interesting_things"

    f4 = f3.text_thing
    assert isinstance(f4, File)
    assert f4.exists()

    f5 = f2 / "red-delicious"
    assert isinstance(f5, File)
    assert f5.exists()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module defines unit tests for `ptutils.cli.simple`.
"""
import pytest
import argparse
from ptutils.cli.simple import (
    _arg_name,
    arg,
    get_command,
    run_command,
    shared_arg,
    use_shared_arg,
    shared_arg_exists,
    remove_shared_arg,
    arg_group,
    use_arg_group,
    arg_group_exists,
    remove_arg_group,
    command,
    command_exists,
    remove_command,
    get_argument,
    has_argument,
    _resolve_command,
    _resolve,
    _walk_argument_specs,
    get_args_parser
)


# ------------------------------------------------------------------------------------------------------------------------
def test_double_decorate_raises():
    try:
        with pytest.raises(Exception) as _:
            @command
            @command
            def my_command():
                pass
    finally:
        remove_command('my_command')
        assert not command_exists('my_command')


# ------------------------------------------------------------------------------------------------------------------------
def test_duplicate_arg_group_raises():
    try:
        @arg_group
        def fruits():
            pass
        assert arg_group_exists('fruits')

        with pytest.raises(Exception) as _:
            @arg_group  # noqa: F811
            def fruits():
                pass

    finally:
        remove_arg_group('fruits')
        assert not arg_group_exists('fruits')


# ------------------------------------------------------------------------------------------------------------------------
def test_mixing_command_and_arg_group_raises():
    try:
        with pytest.raises(Exception) as _:
            @command  # noqa: F811
            @arg_group
            def fruits():
                pass

    finally:
        remove_arg_group('fruits')
        assert not arg_group_exists('fruits')

    try:
        with pytest.raises(Exception) as _:
            @arg_group
            @command
            def my_command():
                pass

    finally:
        remove_command('my_command')
        assert not command_exists('my_command')


# ------------------------------------------------------------------------------------------------------------------------
def test_removing_nonexistent_raises():
    with pytest.raises(Exception) as _:
        remove_shared_arg('DNE')

    with pytest.raises(Exception) as _:
        remove_arg_group('DNE')

    with pytest.raises(Exception) as _:
        remove_command('DNE')

    with pytest.raises(ValueError) as _:
        get_argument('DNE', 'whatever')


# ------------------------------------------------------------------------------------------------------------------------
def test_arg_not_present_returns_none():
    try:
        @command
        def my_command():
            pass

        assert get_argument(my_command, 'DNE') is None
        assert get_argument('my_command', 'DNE') is None

    finally:
        remove_command('my_command')
        assert not command_exists('my_command')

    try:
        @arg_group
        def fruits():
            pass

        assert get_argument(fruits, 'DNE') is None
        assert get_argument('fruits', 'DNE') is None

    finally:
        remove_arg_group(fruits)
        assert not arg_group_exists('fruits')


# ------------------------------------------------------------------------------------------------------------------------
def test_arg_on_undecorated_raises():
    with pytest.raises(Exception) as _:
        @arg('-f', dest='free', action='store_true')
        def undecorated():
            pass


# ------------------------------------------------------------------------------------------------------------------------
def test_use_arg_group_on_undecorated_raises():
    try:
        with pytest.raises(Exception) as _:
            @arg_group
            def fruits():
                pass

            @use_arg_group('fruits')
            def undecorated():
                pass
    finally:
        remove_arg_group('fruits')
        assert not arg_group_exists('fruits')


# ------------------------------------------------------------------------------------------------------------------------
def test_invoking_arg_group_func_raises():
    try:
        @arg_group
        def fruits():
            pass

        with pytest.raises(Exception) as _:
            fruits()

    finally:
        remove_arg_group('fruits')
        assert not arg_group_exists('fruits')


# ------------------------------------------------------------------------------------------------------------------------
def test_shared_arg_func_raises():
    try:
        cherry = shared_arg('-c', '--cherry', dest='cherry')

        with pytest.raises(Exception) as _:
            cherry()

    finally:
        remove_shared_arg('cherry')
        assert not shared_arg_exists(cherry)


# ------------------------------------------------------------------------------------------------------------------------
def test_resolve_none_raises():
    with pytest.raises(TypeError) as _:
        _resolve(None)


# ------------------------------------------------------------------------------------------------------------------------
def test_resolve_command_none_raises():
    with pytest.raises(TypeError) as _:
        _resolve_command(None)


# ------------------------------------------------------------------------------------------------------------------------
@pytest.fixture
def arg_group_fruits():
    @arg("--cherry")
    @arg("--banana")
    @arg("--apple")
    @arg_group
    def fruits():
        pass

    yield fruits

    remove_arg_group('fruits')


# ------------------------------------------------------------------------------------------------------------------------
@pytest.fixture
def arg_group_veggies():
    @arg("--carrots")
    @arg("--broccoli")
    @arg("--asparagus")
    @arg_group
    def veggies():
        pass

    yield veggies

    remove_arg_group('veggies')


# ------------------------------------------------------------------------------------------------------------------------
@pytest.fixture
def arg_group_meats():
    anchovies = shared_arg("--anchovies")
    bacon = shared_arg("--bacon")
    _ = shared_arg("--chicken")

    @use_shared_arg(anchovies)
    @use_shared_arg(bacon)
    @use_shared_arg('chicken')
    @arg_group
    def meats():
        pass

    yield meats

    remove_arg_group('meats')
    remove_shared_arg('anchovies')
    remove_shared_arg('bacon')
    remove_shared_arg('chicken')


# ------------------------------------------------------------------------------------------------------------------------
COOK_CALLED_MARKER = []


# ------------------------------------------------------------------------------------------------------------------------
@pytest.fixture
def hierarchical_cli(arg_group_fruits, arg_group_veggies, arg_group_meats):
    global COOK_CALLED_MARKER
    COOK_CALLED_MARKER = []

    @use_arg_group('fruits')
    @use_arg_group('veggies')
    @use_arg_group('meats')
    @command
    def cook(**kwargs):
        global COOK_CALLED_MARKER
        COOK_CALLED_MARKER.append(kwargs)

    yield cook

    remove_command('cook')
    COOK_CALLED_MARKER = []


# ------------------------------------------------------------------------------------------------------------------------
HIERARCHICAL_CLI_SPECS = dict(
    (name, {"kind": "arg", "name": name, "args": tuple([f"--{name}"]), "kwargs": {}})
    for name in [
        'apple',
        'banana',
        'cherry',
        'asparagus',
        'broccoli',
        'carrots',
        'anchovies',
        'bacon',
        'chicken'
    ]
)


# ------------------------------------------------------------------------------------------------------------------------
def test_args_parser(hierarchical_cli):
    p = get_args_parser(
        program = "program",
        version = "version",
        copyright_message = "(C) Someone, 1984."
    )
    assert isinstance(p, argparse.ArgumentParser)
    args = p.parse_args(
        [
            "cook",
            "--apple", "granny-smith",
            "--banana", "cavendish",
            "--cherry", "bing",

            "--asparagus", "skinny",
            "--broccoli",  "green",
            "--carrots",   "pointy",

            "--anchovies", "sea",
            "--bacon",     "land",
            "--chicken",   "sky"
        ]
    )
    assert args.apple     == "granny-smith"
    assert args.banana    == "cavendish"
    assert args.cherry    == "bing"
    assert args.asparagus == "skinny"
    assert args.broccoli  == "green"
    assert args.carrots   == "pointy"
    assert args.anchovies == "sea"
    assert args.bacon     == "land"
    assert args.chicken   == "sky"

    p2 = get_args_parser(
        program = "dfsadfas",
        version = "fasdfsadf",
        copyright_message = "asdfasdf"
    )
    assert p2 is p


# ------------------------------------------------------------------------------------------------------------------------
def test_run_command(hierarchical_cli):
    global COOK_CALLED_MARKER
    assert len(COOK_CALLED_MARKER) == 0

    rc = run_command(
        args = [
            "cook",
            "--apple", "granny-smith",
            "--banana", "cavendish",
            "--cherry", "bing",

            "--asparagus", "skinny",
            "--broccoli",  "green",
            "--carrots",   "pointy",

            "--anchovies", "sea",
            "--bacon",     "land",
            "--chicken",   "sky"
        ],
        program = "program",
        version = "version",
        copyright_message = "(C) Someone, 1984."
    )
    assert rc == 0
    assert len(COOK_CALLED_MARKER) == 1
    assert isinstance(COOK_CALLED_MARKER[0], dict)
    args = COOK_CALLED_MARKER[0]

    assert args['apple']     == "granny-smith"
    assert args['banana']    == "cavendish"
    assert args['cherry']    == "bing"
    assert args['asparagus'] == "skinny"
    assert args['broccoli']  == "green"
    assert args['carrots']   == "pointy"
    assert args['anchovies'] == "sea"
    assert args['bacon']     == "land"
    assert args['chicken']   == "sky"


# ------------------------------------------------------------------------------------------------------------------------
@pytest.mark.parametrize(
    argnames  = ["var_name", "var_spec"],
    argvalues = HIERARCHICAL_CLI_SPECS.items(),
    ids       = [ f"{k}->{v}" for (k, v) in HIERARCHICAL_CLI_SPECS.items() ]
)
def test_walk_argument_specs(hierarchical_cli, var_name, var_spec):
    assert get_argument(hierarchical_cli, var_name) == var_spec

    for spec in _walk_argument_specs(hierarchical_cli):
        if _resolve(_arg_name) == spec['name']:
            assert spec == var_spec


# ------------------------------------------------------------------------------------------------------------------------
def test_walk_argument_bad_subject_raises():
    with pytest.raises(ValueError) as _:
        for __ in _walk_argument_specs(None):
            pass


# ------------------------------------------------------------------------------------------------------------------------
BAD_CMD_INFO_TYPE_CASES = [
    (None, ), ("None", ), ([], ), (True, ), (False, )
]


# ------------------------------------------------------------------------------------------------------------------------
@pytest.mark.parametrize(
    argnames  = ["cmd_info"],
    argvalues = BAD_CMD_INFO_TYPE_CASES,
    ids       = [ f"{x}" for x in BAD_CMD_INFO_TYPE_CASES ]
)
def test_cmd_info_not_mapping_raises(cmd_info):

    def f_bad_cmd_info():
        pass

    f_bad_cmd_info.cmd_info = None
    with pytest.raises(ValueError) as _:
        for __ in _walk_argument_specs(f_bad_cmd_info):
            pass


# ------------------------------------------------------------------------------------------------------------------------
MALFORMED_CMD_INFO_CASES = [
    ({}, ValueError),                                                     # arg info not present
    ({'arg_info': None}, RuntimeError),                                   # arg info not a mapping or list of mappings
    ({'arg_info': [{}]}, ValueError),                                     # no kind field
    ({'arg_info': [{'kind': 'arg_group'}]}, Exception),                  # no name specified in arg group
    ({'arg_info': [{'kind': 'arg_group', 'name': 'DNE'}]}, Exception),   # arg group name does not exist
]


# ------------------------------------------------------------------------------------------------------------------------
@pytest.mark.parametrize(
    argnames  = ["cmd_info", "exception"],
    argvalues = MALFORMED_CMD_INFO_CASES,
    ids       = [ f"{x}" for x in MALFORMED_CMD_INFO_CASES ]
)
def test_malformed_cmd_info_raises(cmd_info, exception):

    def f_bad_cmd_info():
        pass

    f_bad_cmd_info.__cmd_info__ = cmd_info

    with pytest.raises(exception) as _:
        for __ in _walk_argument_specs(f_bad_cmd_info):
            pass


# ------------------------------------------------------------------------------------------------------------------------
def test_no_arg_positionals_raises():
    try:
        with pytest.raises(Exception) as _:
            @arg()
            @command
            def my_command():
                pass
    finally:
        remove_command('my_command')
        assert not command_exists('my_command')


# ------------------------------------------------------------------------------------------------------------------------
def test_using_unknown_arg_group_raises():
    try:
        with pytest.raises(Exception) as _:
            @use_arg_group('DNE')
            @command
            def my_command():
                pass
    finally:
        remove_command('my_command')
        assert not command_exists('my_command')


# ------------------------------------------------------------------------------------------------------------------------
def test_indeterminate_arg_name_raises():
    try:
        with pytest.raises(Exception) as _:
            @arg()
            @command
            def my_command():
                pass
    finally:
        remove_command('my_command')
        assert not command_exists('my_command')


# ------------------------------------------------------------------------------------------------------------------------
def test_cli_simple_create_command():
    """ Test ability to create a command using the `ptutils.cli.simple.command` decorator. """
    with pytest.raises(Exception) as _:
        get_command('my_command')

    try:
        @command
        def my_command():
            pass
        assert command_exists('my_command')
        assert get_command('my_command') is not None

    finally:
        remove_command('my_command')
        assert not command_exists('my_command')


# ------------------------------------------------------------------------------------------------------------------------
def test_cli_simple_create_arg_group():
    """ Test ability to create an argument group using the `ptutils.cli.simple.arg_group` decorator. """
    try:
        @arg_group
        def fruits():
            pass
        assert arg_group_exists('fruits')
    finally:
        remove_arg_group('fruits')
        assert not arg_group_exists('fruits')


# ------------------------------------------------------------------------------------------------------------------------
def test_cli_simple_create_shared_arg():
    """ Test ability to create a shared argument using the `ptutils.cli.simple.shared_arg` function. """
    try:
        sa = shared_arg(
            '-w', '--whatchamacallits',
            dest = "whatchamacallits"
        )
        assert shared_arg_exists('whatchamacallits')
        assert shared_arg_exists(sa)

    finally:
        remove_shared_arg(sa)
        assert not shared_arg_exists('whatchamacallits')


# ------------------------------------------------------------------------------------------------------------------------
def test_resolve_shared_arg_func_returns_name():
    try:
        sa = shared_arg('-s', '--shared')
        assert sa.__name__ == 'shared'
        assert _resolve(sa) == 'shared'

    finally:
        remove_shared_arg(sa)
        assert not shared_arg_exists('whatchamacallits')


# ------------------------------------------------------------------------------------------------------------------------
def test_cli_simple_use_shared_arg():
    """ Test ability to use a shared argument using the `ptutils.cli.simple.use_shared_arg` function. """
    try:
        # define shared arg
        batch = shared_arg(
            '-b', '--batch',
            metavar = "BATCH_SIZE",
            help = "Batch size"
        )

        # define command using shared arg
        @use_shared_arg('batch')
        @command
        def my_command(batch: int = 5):
            assert batch == 25

        # check
        assert shared_arg_exists('batch')
        assert shared_arg_exists(batch)
        assert command_exists('my_command')
        assert command_exists(my_command)
        assert get_argument('my_command', 'batch') == {
            "kind":   "arg",
            "args":   tuple(['-b', '--batch']),
            "kwargs": {
                "metavar": "BATCH_SIZE",
                "help": "Batch size"
            },
            "name": "batch"
        }
        assert get_argument('my_command', batch) == {
            "kind":   "arg",
            "args":   tuple(['-b', '--batch']),
            "kwargs": {
                "metavar": "BATCH_SIZE",
                "help": "Batch size"
            },
            "name": "batch"
        }
        assert has_argument(my_command, batch)
        assert has_argument('my_command', 'batch')

    finally:
        # cleanup
        remove_command('my_command')
        remove_shared_arg(batch)

        assert not command_exists('my_command')
        assert not shared_arg_exists('batch')


# ------------------------------------------------------------------------------------------------------------------------
def test_cli_simple():
    """ Tests many functions of the simple cli system together. """
    try:
        # define a fruits arguments group
        @arg(
            '-b', '--banana',
            help    = 'Yummy yellow fruit',
            dest    = 'banana',
            metavar = 'BANANA',
            default = 'banana'
        )
        @arg(
            '-c', '--cherry',
            help    = 'Yummy red fruit',
            dest    = 'cherry',
            metavar = 'CHERRY',
            default = 'cherry'
        )
        @arg_group
        def fruits():
            pass

        # define a vegetables arguments group
        @arg(
            '-s', '--squash',
            help    = 'Yucky green vegetable',
            dest    = 'squash',
            metavar = 'SQUASH',
            default = 'squash'
        )
        @arg(
            '-t', '--tomato',
            help    = 'Yummy red vegetable',
            dest    = 'tomato',
            metavar = 'TOMATO',
            default = 'tomato'
        )
        @arg_group
        def vegetables():
            pass

        # define a produce arguments group
        @use_arg_group('fruits')
        @use_arg_group('vegetables')
        @arg_group
        def produce():
            pass

        # define a shared argument
        shared_arg(
            '-w', '--whatchamacallits',
            dest = "whatchamacallits"
        )

        # define a command using shared argument, argument group, and some directly defined arguments
        @use_shared_arg('whatchamacallits')
        @use_arg_group('produce')
        @arg(
            '-p', '--prefix',
            help    = 'Folder prefix',
            dest    = 'prefix',
            metavar = 'FOLDER',
            default = '/tmp'
        )
        @arg(
            '-f', '--force',
            help    = 'When present, ignore cached results and perform all actions again.',
            dest    = 'force',
            action  = 'store_true',
            default = False
        )
        @command
        def my_command(prefix: str, force: bool, **kwargs) -> None:
            print("I'm now running stuff!")

        assert arg_group_exists('fruits')
        assert has_argument('fruits', 'cherry')
        assert has_argument('fruits', 'banana')

        assert arg_group_exists('vegetables')
        assert has_argument('vegetables', 'squash')
        assert has_argument('vegetables', 'tomato')

        assert arg_group_exists('produce')
        assert has_argument(produce, 'cherry')
        assert has_argument('produce', 'banana')
        assert get_argument(produce, 'squash') is not None
        assert get_argument('produce', 'tomato') is not None

        assert shared_arg_exists('whatchamacallits')
        assert command_exists('my_command')

        assert has_argument('my_command', 'whatchamacallits')
        assert has_argument('my_command', 'cherry')
        assert has_argument('my_command', 'banana')
        assert has_argument('my_command', 'squash')
        assert has_argument('my_command', 'tomato')
        assert has_argument('my_command', 'force')
        assert has_argument('my_command', 'prefix')

        assert get_argument('my_command', 'whatchamacallits') == {
            "kind":   "arg",
            "args":   tuple(['-w', '--whatchamacallits']),
            "kwargs": {
                "dest": "whatchamacallits"
            },
            "name": "whatchamacallits"
        }

    finally:
        remove_shared_arg('whatchamacallits')
        remove_arg_group('produce')
        remove_arg_group('vegetables')
        remove_arg_group('fruits')
        remove_command('my_command')

        assert not command_exists('my_command')
        assert not arg_group_exists('produce')
        assert not arg_group_exists('vegetables')
        assert not arg_group_exists('fruits')
        assert not shared_arg_exists('whatchamacallits')



#!/bin/false
# -*- coding: utf-8 -*-

""" Logging functions. """

# ------------------------------------------------------------------------------------------------------------------------
# Imports
# ------------------------------------------------------------------------------------------------------------------------
import logging
from typing import Dict, Tuple, Union
from ptutils.undefined import default_if_undefined, UNDEFINED, Undefined

# ------------------------------------------------------------------------------------------------------------------------
# Logger config
# ------------------------------------------------------------------------------------------------------------------------
""" The default verbosity to use when not otherwise specified."""
DEFAULT_VERBOSITY = 3


""" The currently configured verbosity. use `get_verbosity` or `set_verbosity` to modify. """
__VERBOSITY__ = DEFAULT_VERBOSITY


# ------------------------------------------------------------------------------------------------------------------------
""" The currently configured log level. use `get_verbosity` or `set_verbosity` to modify. """
__LEVEL__ = logging.INFO


# ------------------------------------------------------------------------------------------------------------------------
""" The currently configured log level name. use `get_verbosity` or `set_verbosity` to modify. """
__LEVEL_NAME__ = 'INFO'


# ------------------------------------------------------------------------------------------------------------------------
""" A mapping between verbosity levels and standard python logging levels. """
__VERBOSITY_LEVELS__: Dict[int, Tuple[int, str]] = {
    5: (logging.NOTSET + 1, "VERBOSE"),
    4: (logging.DEBUG,      "DEBUG"),
    3: (logging.INFO,       "INFO"),
    2: (logging.WARNING,    "WARNING"),
    1: (logging.ERROR,      "ERROR"),
    0: (logging.CRITICAL,   "CRITICAL")
}


# ------------------------------------------------------------------------------------------------------------------------
def clip_verbosity_level(verbosity: int) -> int:  # pragma: no cover
    """
    Given a verbosity level, clip it to the allowable range of [0,5] (inclusive).

    Parameters
    ----------
    verbosity : int
        An integer verbosity level.

    Returns
    -------
    int
        The value clipped to the range [0,5].

    Raises
    ------
    Exception
        When verbosity is not or can not be converted to an integer.

    """
    verbosity = int(verbosity)
    if verbosity < 0:
        return 0
    if verbosity > 5:
        return 5
    return verbosity


# ------------------------------------------------------------------------------------------------------------------------
def translate_verbosity_level(verbosity: int) -> Tuple[int, str]:  # pragma: no cover
    """
    Given a verbosity level, get the python `logging` standard library level and level name.

    Parameters
    ----------
    verbosity : int
        A verbosity level on the range [0,5].

    Returns
    -------
    Tuple[int, str]
        The corresponding logging level and level name.

    Raises
    ------
    ValueError
        If verbosity is not an integer on the range [0,5].
    """
    if verbosity in __VERBOSITY_LEVELS__:
        return __VERBOSITY_LEVELS__[verbosity]
    else:
        raise ValueError(f"I don't know how to interpret verbosity level {verbosity}.")


# ------------------------------------------------------------------------------------------------------------------------
def set_verbosity(verbosity: Union[int, Undefined] = UNDEFINED) -> None:  # pragma: no cover
    """
    Set the logging verbosity.

    Parameters
    ----------
    verbosity : Union[int, Undefined], optional
        A verbosity level on the range [0,5] or UNDEFINED, by default UNDEFINED. When
        UNDEFINED, the default value of `DEFAULT_VERBOSITY` is used.

    Notes
    -----
    `verbosity` may be any thing convertaible to an integer, and will be clipped to
    the valid range use.
    """
    global __VERBOSITY__
    global __LEVEL__
    global __LEVEL_NAME__

    # set the verbosity to default if not specified.
    verbosity = default_if_undefined(verbosity, DEFAULT_VERBOSITY)

    # clip verbosity to allowed range.
    verbosity = clip_verbosity_level(verbosity)

    # translate verbosity to log level
    level, name = translate_verbosity_level(verbosity)

    # apply verbosity level
    logger.setLevel(level)
    __VERBOSITY__  = verbosity
    __LEVEL__      = level
    __LEVEL_NAME__ = name


# ------------------------------------------------------------------------------------------------------------------------
def get_verbosity() -> int:  # pragma: no cover
    """
    Get the currently configured logging verbosity.

    Returns
    -------
    int
        The verbosity value on range [0,5].
    """
    return __VERBOSITY__


# ------------------------------------------------------------------------------------------------------------------------
def get_level() -> int:  # pragma: no cover
    """
    Get the currently configured logging level.

    Returns
    -------
    int
        The verbosity value on range [0,5].
    """
    return __LEVEL__


# ------------------------------------------------------------------------------------------------------------------------
def get_level_name() -> str:  # pragma: no cover
    """
    Get the currently configured logging level name.

    Returns
    -------
    str
        One of VERBOSE, DEBUG, INFO, WARNING, ERROR, or CRITICAL
    """
    return __LEVEL_NAME__


# ------------------------------------------------------------------------------------------------------------------------
def getLogger(*args, **kwargs) -> logging.Logger:  # pragma: no cover
    """
    Get a logger.

    Returns
    -------
    logging.Logger
        a logger.
    """
    return logging.getLogger(*args, **kwargs)


# ------------------------------------------------------------------------------------------------------------------------
# Root logger initialization
# ------------------------------------------------------------------------------------------------------------------------
logging.basicConfig(level=__LEVEL__)
logger = logging.getLogger()
logger.setLevel( __LEVEL__ )
